
#include <algorithm>
#include <cinttypes>
#include <climits>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <malloc.h>
#include <ogcsys.h>
#include <gccore.h>
#include <unistd.h>



constexpr size_t PAGE_SIZE = 4096;

static volatile u32* const s_pi_reg = reinterpret_cast<volatile u32*>(0xCC003000);

static volatile bool s_pi_error_occurred = false;

static void *xfb = NULL;
static GXRModeObj *rmode = NULL;

void *Initialise();

static int tests_total = 0;
static int tests_passed = 0;

#define DO_TEST(condition, fail_msg, ...) \
  do { \
    tests_total++; \
    if (!condition) { \
      printf("test %i failed: " fail_msg, tests_total, ##__VA_ARGS__); \
      usleep(100000); /* scroll slightly slower */ \
    } else { \
      tests_passed++; \
    } \
  } while (0)


static u32 Read(volatile u8* ptr)
{
  return *reinterpret_cast<volatile u32*>(ptr);
}

static void Write(volatile u8* ptr, u32 value, u32 size)
{
  switch (size)
  {
  case 4:
    asm volatile("stw %0, 0(%1)" :: "r"(value), "r"(ptr));
    break;
  case 2:
    asm volatile("sth %0, 0(%1)" :: "r"(value), "r"(ptr));
    break;
  case 1:
    asm volatile("stb %0, 0(%1)" :: "r"(value), "r"(ptr));
    break;
  }
}

namespace Common {
template <typename T>
constexpr size_t BitSize() noexcept
{
  return sizeof(T) * CHAR_BIT;
}

template <typename T>
constexpr T RotateRight(const T value, size_t amount) noexcept
{
  static_assert(std::is_unsigned<T>(), "Can only rotate unsigned types right.");

  amount %= BitSize<T>();

  if (amount == 0)
    return value;

  return static_cast<T>((value >> amount) | (value << (BitSize<T>() - amount)));
}
}


static void WriteWithSimulatedQuirks(uintptr_t alignment, volatile u8* ptr, u32 value, u32 size)
{
  const uintptr_t misalignment_32 = alignment & 3;

  if (misalignment_32 || size < 4)
  {
    const uintptr_t misalignment_64 = alignment & 7;
    const uintptr_t count = misalignment_64 + size;

    value = Common::RotateRight(value, ((misalignment_32 + size) & 3) * 8);

    for (uintptr_t i = 0; i < count; i += 8)
    {
      Write(ptr - misalignment_64 + i, value, sizeof(u32));
      Write(ptr - misalignment_64 + i + 4, value, sizeof(u32));
    }
  }
  else
  {
    Write(ptr, value, size);
  }
}

static void UnalignedStoresTest(volatile u8* ptr, u32 size, bool cached_memory)
{
  printf("Starting test using ptr 0x%" PRIxPTR ", size %u\n",
                 reinterpret_cast<uintptr_t>(ptr), size);

  volatile u8 reference_buffer[32];
  const u32 word = 0x12345678;

  for (size_t i = 0; i <= 28; ++i)
  {
    std::fill(ptr, ptr + 32, 0);
    std::fill(reference_buffer, reference_buffer + 32, 0);

    s_pi_error_occurred = false;

    // The actual write that we are testing the behavior of
    Write(ptr + i, word, size);

    usleep(10000);

    if (cached_memory)
      Write(reference_buffer + i, word, size);
    else
      WriteWithSimulatedQuirks(i, reference_buffer + i, word, size);

    DO_TEST(std::equal(ptr, ptr + 32, reference_buffer),
            "%u-byte write to 0x%" PRIxPTR " failed\n"
            "ACTUAL:   %08X%08X %08X%08X %08X%08X %08X%08X\n"
            "EXPECTED: %08X%08X %08X%08X %08X%08X %08X%08X\n",
            size, reinterpret_cast<uintptr_t>(ptr + i),
            Read(ptr), Read(ptr + 4),
            Read(ptr + 8), Read(ptr + 12),
            Read(ptr + 16), Read(ptr + 20),
            Read(ptr + 24), Read(ptr + 28),
            Read(reference_buffer), Read(reference_buffer + 4),
            Read(reference_buffer + 8), Read(reference_buffer + 12),
            Read(reference_buffer + 16), Read(reference_buffer + 20),
            Read(reference_buffer + 24), Read(reference_buffer + 28));

    // The pattern behind when PI ERROR interrupts are generated is not understood.
    // When pi_error_expected == false, a real Wii does not generate an interrupt, but when
    // pi_error_expected == true it may or may not. Disable this part of the test for now.
    /*
    int pi_error_expected = (i & 3) || size < 4;
    int pi_error_actual = s_pi_error_occurred;
    DO_TEST(pi_error_actual == pi_error_expected,
            "%u-byte write to 0x%" PRIxPTR " failed\n"
            "ACTUAL:              EXPECTED:\n"
            "pi_error_occurred=%d  pi_error_occurred=%d\n",
            size, reinterpret_cast<uintptr_t>(ptr + i),
            pi_error_actual, pi_error_expected);
    */
  }
}


static void PIErrorHandler(u32 nIrq, void* pCtx)
{
  // Clear the interrupt
  s_pi_reg[0] = 0x01;

  s_pi_error_occurred = true;
}

int main(int argc, char **argv) {

	xfb = Initialise();

	IRQ_Request(IRQ_PI_ERROR, PIErrorHandler, nullptr);
  __UnmaskIrq(IM_PI_ERROR);

  // Get a pointer to a page boundary, with at least 32 bytes available on each side of the boundary
  printf("Allocating memory...\n");
  u8* memory_allocation = new u8[PAGE_SIZE + 32 * 4];

  volatile u8* cached_ptr = reinterpret_cast<volatile u8*>(
          reinterpret_cast<uintptr_t>(memory_allocation - 32) & ~(PAGE_SIZE - 1));

  UnalignedStoresTest(cached_ptr - 16, 1, true);
  UnalignedStoresTest(cached_ptr - 16, 2, true);
  UnalignedStoresTest(cached_ptr - 16, 4, true);

  printf("Invalidating cache...\n");
  asm volatile("dcbi %0, %1" :: "r"(cached_ptr), "r"(-32));
  asm volatile("dcbi %0, %1" :: "r"(cached_ptr), "r"(0));

  volatile u8* uncached_ptr = reinterpret_cast<volatile u8*>(MEM_K0_TO_K1(cached_ptr));

  UnalignedStoresTest(uncached_ptr - 16, 1, false);
  UnalignedStoresTest(uncached_ptr - 16, 2, false);
  UnalignedStoresTest(uncached_ptr - 16, 4, false);

  delete[] memory_allocation;

  printf("Passed %i of %i tests\n", tests_passed, tests_total);

	while(1) {

		VIDEO_WaitVSync();
		PAD_ScanPads();

		int buttonsDown = PAD_ButtonsDown(0);

		if (buttonsDown & PAD_BUTTON_START) {
			exit(0);
		}
	}

	return 0;
}

void * Initialise() {

	void *framebuffer;

	VIDEO_Init();
	PAD_Init();

	rmode = VIDEO_GetPreferredMode(NULL);

	framebuffer = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
	console_init(framebuffer,20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);

	VIDEO_Configure(rmode);
	VIDEO_SetNextFramebuffer(framebuffer);
	VIDEO_SetBlack(FALSE);
	VIDEO_Flush();
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE) VIDEO_WaitVSync();

	return framebuffer;

}
